from __future__ import print_function, division, unicode_literals, absolute_import
import logging

__title__ = 'actappliance'
__version__ = '0.1.11'

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())


# Hoist appliance up
from actappliance.appliance import Appliance  # NOQA


# from urllib3
def add_stderr_logger(level=logging.DEBUG):
    """
    Helper for quickly adding a StreamHandler to the logger. Useful for
    debugging.
    Returns the handler after adding it.
    """
    # This method needs to be in this __init__.py to get the __name__ correct
    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.debug('Added a stderr logging handler to logger: %s', __name__)
    return handler


# Clean up
del NullHandler


def add_generic_logger(file_level=logging.DEBUG):
    logging.basicConfig(filename='actappliance.out',
                        level=file_level,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M',)

    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger(__name__).addHandler(console)



#!/usr/bin/env python
from __future__ import print_function, division, unicode_literals, absolute_import
# from future.utils import with_metaclass

import logging
from actappliance.act_errors import act_errors, ACTError
from actappliance.permissions import AppliancePermissions
from actappliance.connections import ApplianceSsh, ApplianceRest


appl_types = ['cds', 'sky', 'psky']
appl_types_str = ', '.join(appl_types[:-1]) + ', or ' + appl_types[-1]


class Appliance(object):
    """
    An appliance represents CDS or SKY actifio product.

    As a general rule use cmd method for all udsinfo and udstask commands. This allows them to be changed dynamically
    at runtime. If you are sending a hidden rest command use rest and if you are sending a command that can't be sent
    over rest use raw.

    Usage:
     a = Appliance(ip_address='172.27.4.0'
     r = a.cmd('udsinfo lshost -filtervalue hostname=dog&hosttype=puppey')
     hostid = r.parse(k='id')

    Commands themselves should go into Libraries not this class.
    """

    def __init__(self, hostname=None, ip_address=None, connection_type='rest', appliance_type=None, version=None,
                 ssh_args=None, rest_params=None):
        """
        Setup config info as well as permissions

        ssh_args should normally have username, password, port as keys.
        rest_params should normally have name, password, vendorkey as keys.

        :param hostname: DNS name of machine to connect to (hostname or ip_address required)
        :param ip_address: Ip address of machine to connect to (hostname or ip_address required)
        :param connection_type: Default way to connect to appliance (options: cli, rest)
        :param appliance_type: Category of appliance
        :param ssh_args: Dictionary of ssh connection args to send to paramiko's connect method
        :param rest_params: Dictionary of connection params to send alongside RESTful requests
        """
        # These permissions are to stop users from hitting unexpected failures. Not for security.
        self.perms = AppliancePermissions()

        self.logger = logging.getLogger(__name__)

        # Get system info
        self.hostname = hostname
        self.ip_address = ip_address
        self.system = ip_address or hostname
        if not self.system:
            self.logger.warn('Ip address or hostname required to send commands')

        # Get connection info
        self.ssh_args = ssh_args if ssh_args is not None else {}
        self.rest_params = rest_params if rest_params is not None else {}

        # Get optional appliance type
        self._appliance_type = appliance_type if appliance_type is not None and appliance_type.lower() in appl_types \
            else None
        self.version = version

        # Ensure connection input is valid
        connection_type = connection_type.lower()
        if connection_type in ('rest', 'cli'):
            self.connection_type = connection_type
        else:
            raise ValueError("connection_type accepts 'rest' or 'cli'; {} was given".format(connection_type))

        # Make connections for commands
        self._make_connections()

    def _make_connections(self):
        self.rest_conn = ApplianceRest(self.system, self.rest_params)
        self.ssh_conn = ApplianceSsh(self.system, self.ssh_args)

    def cmd(self, operation, **update_cmds):
        """
        Connect to the machine using the classes connection type and send the given command.

        If an argument needs to be sent in the operation and not in kwargs with argument make sure it is first in the
        series of commands.

        :param operation: CLI-like command to send
        :param update_cmds: add or overwrite commands in the operation. Will not add if value is none.
        :return: ActResponse object
        """
        self.perms.has_cmd()
        if self.connection_type == "rest":
            return self.rest(operation, **update_cmds)
        elif self.connection_type == "cli":
            return self.cli(operation, **update_cmds)
        else:
            raise RuntimeError("Couldn't find suitable cmd method based on connection type. {} "
                               "provided".format(self.connection_type))

    def rest(self, operation, **update_cmds):
        """
        Use when you explicitly want to connect over RESTful API.

        :param operation: CLI-like command to be sent. ex. udsinfo lshost
        :param update_cmds: keyword arguments to be sent with command
        :return:
        """
        self.perms.has_rest()

        return self.rest_conn.cmd(operation, **update_cmds)

    def cli(self, operation, **update_cmds):
        """
        Use when you explicitly want to connect over ssh.

        :param operation: CLI-like command to be sent. ex. udsinfo lshost
        :param update_cmds: keyword arguments to be sent with command
        :return: ActResponse object
        """
        self.perms.has_ssh()

        return self.ssh_conn.cmd(operation, **update_cmds)

    def raw(self, command, *args, **kwargs):
        """
        Use when you explicitly want to connect over ssh and send raw ssh commands.

        This houses the functionality unlike rest method because I want to avoid passing around the client to retain
        channels.
        :param command:
        :return: stdout, stderr, exit status
        """
        self.perms.has_ssh()

        return self.ssh_conn.raw(command, *args, **kwargs)

    def ui_call(self, method, url_suffix, params=None, data=None, **kwargs):
        """
        Used to send Actifio desktop rest calls.

        :param method: type of rest call to make. ie. GET, POST
        :param url_suffix: What follows actifio in the UI call. i.e. https://<some_hostname>/actifio/<url_suffix here>
        :return:
        """
        self.perms.has_rest()
        return self.rest_conn.ui_call(method, url_suffix, params, data, **kwargs)

    def teardown(self):
        self.rest_conn.disconnect()
        self.ssh_conn.disconnect()

    @property
    def appliance_type(self):
        if not self._appliance_type:
            self.get_appliance_type()
        return self._appliance_type

    def get_appliance_type(self):
        """Find the appliance's type. ex. sky, cds"""
        from paramiko import AuthenticationException

        appliance_type = None
        try:
            self.logger.debug("Searching for appliance type over ssh.")
            out, err, rc = self.raw('rpm -qa | grep actsys | grep -ho -e svc -e psky -e sky', statuses=[0])
            try:
                actsys_type = out[0].lower()
            except IndexError:
                raise ValueError("rpm actsys didn't return a meaningful type.")
            appliance_type = actsys_type.replace('svc', 'cds')
            if appliance_type not in appl_types:
                msg = "Appliance type found from rpm -qa isn't {0}".format(appl_types_str)
                self.logger.exception(msg)
                raise ValueError(msg)
        except (AuthenticationException, ValueError):
            self.logger.debug("Failed to get appliance type over ssh, searching with RESTful API.")
            r = self.rest('udsinfo lsversion')
            r.raise_for_error()
            for di in r['result']:
                for v in di.values():
                    if v.lower() in appl_types:
                        appliance_type = v.lower()
                        break

        assert appliance_type in appl_types, "Invalid appliance type {0} found".format(appliance_type)
        self.logger.info("Successfully got appliance type {}.".format(appliance_type))
        self._appliance_type = str(appliance_type)

    def get_version(self):
        r = self.cmd('udsinfo lsversion')
        self.version = r.parse(k='version', m_k='component', m_v='srv-revision')
        if not self.version:
            logging.debug('This system looks older and may call srv-revision psrv-revision, attempting to grab.')
            self.version = r.parse(k='version', m_k='component', m_v='psrv-revision')
        logging.info('srv-version is {}'.format(self.version))

    def new_user(self, new_name, new_password=None):
        """
        This is an idempotent way to select and sign in as a new user with default rights.

        :param new_name: New username to create or find
        :param new_password: New password to set or use
        :return: New Appliance object with the new user information based off of the current instance
        """
        if new_password is None:
            new_password = self.rest_params.get('password')

        # we don't need to check_login if exceptions aren't hit
        check_login = False
        try:
            r = self.cmd('udstask mkuser', name=new_name, password=new_password)
            r.raise_for_error()
        except act_errors[10006]:
            # This assumes the password of the created user is new_password
            logging.info('User {} already existed.'.format(new_name))
            # Check the password we have works
            check_login = True
        finally:
            self.teardown()

        # Remove the ssh permissions
        new_perms = self.perms
        new_perms.ssh = False
        # set the new params
        new_rest_params = self.rest_params.copy()
        new_rest_params['name'] = new_name
        new_rest_params['password'] = new_password

        new_args = []
        for arg in [self.hostname, self.ip_address, self.connection_type, self.appliance_type, self.version,
                    self.ssh_args]:
            try:
                new_args.append(arg.copy())
            except AttributeError:
                new_args.append(arg)

        new_appliance = Appliance(*new_args, rest_params=new_rest_params)
        new_appliance.perms = new_perms

        # If we found an existing user, make sure our password works
        if check_login:
            r = new_appliance.rest('udsinfo lsversion')
            try:
                r.raise_for_error()
            except ACTError:
                raise RuntimeError("User {} already existed and the provided password was incorrect.".format(new_name))

        return new_appliance



"""Miscellaneous utilities"""
import re

# Following license is for camelize.
# https://github.com/jpvanhal/inflection

# Copyright (C) 2012-2015 Janne Vanhala
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def camelize(string, uppercase_first_letter=True):
    """
    Convert strings to CamelCase.
    Examples::
        >>> camelize("device_type")
        "DeviceType"
        >>> camelize("device_type", False)
        "deviceType"
    :func:`camelize` can be though as a inverse of :func:`underscore`, although
    there are some cases where that does not hold::
        >>> camelize(underscore("IOError"))
        "IoError"
    :param uppercase_first_letter: if set to `True` :func:`camelize` converts
        strings to UpperCamelCase. If set to `False` :func:`camelize` produces
        lowerCamelCase. Defaults to `True`.
    """
    if uppercase_first_letter:
        return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)
    else:
        return string[0].lower() + camelize(string)[1:]



import shlex
import logging
from concurrent.futures import ThreadPoolExecutor as Executor
from concurrent.futures import TimeoutError
from actappliance.models import ActResponse


def call_timeout_handler(signum, frame):
    """Setup with signal.signal(signal.SIGALRM, call_timeout_handler)"""
    logging.info("Call appears to be hanging, killing call.")
    raise RuntimeError("Call took longer than the timeout. Killing call.")


class ApplianceConnection(object):

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.timeout = 6 * 60 * 60

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError

    def prep_cmd(self, operation, **update_cmds):
        """
        Do what needs to be done before sending the command

        :param operation:
        :param update_cmds:
        :return: A ActCmd object from this class
        """
        return self.ActCmd(operation, **update_cmds)

    def call(self, command_obj):
        """Call that gets the unfiltered connections response"""
        raise NotImplementedError

    def _timeout(self, fnc, *args, **kwargs):
        with Executor(max_workers=1) as p:
            f = p.submit(fnc, *args, **kwargs)
            try:
                return f.result(timeout=self.timeout)
            except TimeoutError:
                logging.exception("Call failed:\n"
                                  "function: {0}\n"
                                  "args: {1}\n"
                                  "kwargs: {2}\n".format(fnc, args, kwargs))
                raise

    def timeout_call(self, command_obj):
        return self._timeout(self.call, command_obj)

    def cmd(self, operation, **update_cmds):
        """Command that makes a call with standardized input and output so it's the same across connections"""
        command = self.prep_cmd(operation, **update_cmds)
        self.timeout_call(command)
        return self.post_cmd(command)

    def post_cmd(self, command_obj):
        """Perform post-cmd handlers and wrap result"""
        # Remove result so we don't save it twice
        result = command_obj.result
        command_obj.result = None
        return ActResponse(result, actifio_command=command_obj)

    def append_filtervalue(self, command, **update_cmds):
        """
        Takes a operation and an update_cmds dict and adds the new filtervalues to the existing one instead of
        overwriting.

        :param command:
        :param update_cmds:
        :return: operation and correctly appended update_cmds
        """
        # handle no input
        if 'filtervalue' not in update_cmds:
            # return empty string so that update will function without TypeError
            return ''
        filtervalue_append = update_cmds.pop('filtervalue')
        if filtervalue_append:
            s_command = shlex.split(command)
            for word in enumerate(s_command):
                if word[1] == '-filtervalue':
                    position = word[0]
                    update_cmds['filtervalue'] = "{}&{}".format(s_command[position + 1], filtervalue_append)
            try:
                update_cmds['filtervalue']
            except KeyError:
                self.logger.warning('No existing filtervalue found returning update_cmd filtervalue only')
                update_cmds['filtervalue'] = filtervalue_append
        return update_cmds

    class ActCmd(object):
        """Class to house attributes for a single cmd"""

        def __init__(self, operation=None, **update_cmds):
            self.operation = operation
            self.update_cmds = update_cmds
            self.call_result = None  # This holds result before it's processed in post_cmd if required
            self.result = None  # This is the temporary holder for the REST-ful response

        def __repr__(self):
            """This will be None if post_cmd gets called, check the ActResponse dictionary for the values"""
            return str(self.result)



from bitpermissions import Permissions


class AppliancePermissions(Permissions):
    """Simple permissions for the Appliance class."""

    # Default to all permissions
    def __init__(self, perms=None, perm_digits=-1):
        if perms is None:
            perms = ['cmd', 'rest', 'ssh']
        super(AppliancePermissions, self).__init__(perms, perm_digits)



import requests
import json
import shlex
import atexit
import logging
import time

from actappliance.models import ActResponse
from actappliance.connections.connection import ApplianceConnection

# suppress warnings from verify=False
from requests.packages.urllib3.exceptions import SecurityWarning
requests.packages.urllib3.disable_warnings(SecurityWarning)
# This is required if we want to support old versions of py2
# from requests.packages.urllib3.exceptions import SNIMissingWarning
# requests.packages.urllib3.disable_warnings(SNIMissingWarning)


class ApplianceRest(ApplianceConnection):

    def __init__(self, system, params, call_timeout=6 * 60 * 60):
        """
        :param system: Appliance to connect to (ip or dns name)
        :param params: Dictionary of params to send to connect; ex. {'name': 'ex', 'password': 'ex', 'vendorkey': 'ex'}
        :param call_timeout: Maximum time to spend attempting an individual call
        """
        super(ApplianceRest, self).__init__()
        self.logger = logging.getLogger(__name__)

        self.system = system
        self.timeout = call_timeout
        self.params = params if params is not None else {}
        self.params.setdefault('vendorkey', 'novend')

        # Set in method 'connect'
        self.s = self.sid = None

    def connect(self):
        """Connect and get a sessionid from the appliance."""
        # Ensure we have everything we need to connect before attempting
        for input_ in ['name', 'password', 'vendorkey']:
            if input_ not in self.params:
                raise ValueError("Missing parameter {}, cannot login without it.".format(input_))

        self.s = requests.Session()

        self.logger.info("Logging in as user {}.".format(self.params.get('name', '<no name found>')))

        result = self.s.post("https://" + self.system + "/actifio/api/login", verify=False, params=self.params)
        result.raise_for_status()
        if 'errorcode' in result.json():
            raise ValueError("Failed to connect: {}".format(result.json()))

        self.sid = result.json()['sessionid']
        self.logger.info('Sessionid {} acquired.'.format(self.sid))

        self.s.keep_alive = False

        # Sign out after execution
        atexit.register(self.disconnect, sid=self.sid)

    def disconnect(self, sid=None):
        """Logout of an actifio rest session."""
        if sid is None:
            sid = self.sid

        # Only try to disconnect if we have a sid. It takes a long time to retry connection to a non-existent machine.
        if sid:
            self.logger.debug("Logging out of session {}".format(sid))
            url = ("https://{}/actifio/api/logout?sessionid={}".format(self.system, sid))
            try:
                self.s.post(url, verify=False)
            except requests.ConnectionError:
                # Don't raise connection error. It's safe to let it timeout.
                logging.warning("Unable to logout of session, it may have already expired.")
            self.s.close()

        # set sid to None for the case where it was set as self.sid
        sid = None

    def prep_cmd(self, command, **update_cmds):
        if not self.sid:
            self.connect()

        command = self.ActCmd(command, **update_cmds)
        command.method, command.url, command.params = self.cli_to_rest(command.operation)
        command.params.update(update_cmds)
        self.logger.debug('Parameters for rest call are {}'.format(command.params))
        return command

    def call(self, command_obj):
        r = self._protect_sid(command_obj.method, command_obj.url, command_obj.params)

        try:
            r.raise_for_status()
        except requests.HTTPError as e:  # NOQA: keep this object for debugging
            self.logger.debug("Bad HTTP Status {}.\nResponse: {}\nUrl {}".format(r.status_code, r.json(), r.url))
            # Let the user handle errors
        command_obj.result = r.json()

    def _protect_sid(self, method, url, params, verify=False, **kwargs):
        """Resend the operation if it looks like the sessionid has been dropped (default times out at 30 minutes)."""
        try:
            r = self.s.request(method, url, params=params, verify=verify, **kwargs)
            self.logger.info('Request url: {}'.format(r.url))
            self.logger.debug(r.content)
            r.raise_for_status()
        except requests.HTTPError as e:
            # errorcode 10020 is User is not logged in
            # {'errorcode': 10020, 'errormessage': 'User is not logged in'}
            try:
                errorcode = r.json()['errorcode']
            except KeyError:
                print(r.json())
                logging.exception("Status code was bad, but no errorcode could be found!")
                raise
            if e.response.status_code == 500 and errorcode == 10020:
                self.logger.exception('Looks like the sessionid expired, re-authenticating and sending again.')
                self.logger.debug(r.json())
                self.logger.debug('Failed command sessionid is {}'.format(self.sid))
                self.connect()
                params['sessionid'] = self.sid
                self.logger.debug('New sessionid is {}'.format(self.sid))
                r = self.s.request(method, url, params=params, verify=verify, **kwargs)

            if e.response.status_code == 500 and errorcode == 10023:
                self.logger.exception('Looks like the server timed out sending again.')
                # TODO: restructure with call class so that reauthentication can be handled at the same time.
                time.sleep(1)
                r = self.s.request(method, url, params=params, verify=verify, **kwargs)

            # Don't re-raise let the user handle all other errors

        return r

    def ui_call(self, method, url_suffix, params=None, data=None, **kwargs):
        """
        Used to send Actifio desktop rest calls.

        :param method: type of rest call to make. ie. GET, POST
        :param input_dictionary: python object to be converted to json
        :param url_suffix: What follows actifio in the UI call. i.e. https://<some_hostname>/actifio/<url_suffix here>
        :return: ActResponse object of request
        """
        if not self.sid:
            self.connect()

        if params is None:
            params = {}
        url = "https://{}/actifio/{}".format(self.system, url_suffix)
        params['sessionid'] = self.sid
        r = self._protect_sid(method, url, params=params, data=data, **kwargs)
        r.raise_for_status()
        # strip the unparseable parts of the response that the UI framework needs and convert to dictionary
        # answer = ast.literal_eval(r.text[1:-1])
        answer = json.loads(r.text[1:-1])
        logging.debug(answer)
        if answer.get('status') != 'ok':
            ar = ActResponse(answer)
            ar.raise_for_error()
        return answer

    def cli_to_rest(self, operation):
        """
        Turn a cli command into rest method, url, and params. Argument should be the first parameter.

        Method and url return are strings while params is a dictionary.
        Argument should be the first command because barring documenting and updating all valueless keynames (i.e.
        -force -nowait etc.) this encoder cannot figure out what parameters are switches and which expect a value.
        :param operation: The full operation to be converted. ex. "udsinfo lshost 1234 -filtervalue some=thing&also=this
        :return: method (str), url (str), params (dict)
        """
        # derive method and call-type from first word
        self.logger.info('Converting cli command "{0}" to rest.'.format(operation))
        command = operation.partition(' ')[0]
        if command == "udsinfo":
            method = 'get'
            call = 'info'
        elif command == "udstask":
            method = 'post'
            call = 'task'
        elif command == "sainfo":
            method = 'get'
            call = 'shinfo'
        else:
            raise ValueError("Actifio appliance doesn't support operation {}, accepted values "
                             "(udsinfo, udstask, sainfo)".format(command))

        # separate action (ex. "lsbackup") and params (ex. "1234 -force -filtervalue key=value") from body
        body = operation.partition(' ')[2]
        split_body = shlex.split(body)
        try:
            action = split_body[0]
        except IndexError as e:
            raise IndexError('Operation provided to convert to cli has no action. ex. "mkhost". {}'.format(e))
        chaos_params = split_body[1:]

        # create param dictionary
        params = {}
        k = None
        v = None
        for i in reversed(chaos_params):
            if i.startswith('-'):
                k = i[1:]
            else:
                if v is None:
                    v = i
                else:
                    # handle sequential values
                    params['argument'] = v
                    v = i
            if k:
                if v:
                    params[k] = v
                    k = None
                    v = None
                else:
                    params[k] = 'true'
                    k = None
                    # don't raise not implemented here, let appliance tell user input problems
        # grab the argument if it's the first parameter
        if v is not None:
            params['argument'] = v

        if self.sid:
            params['sessionid'] = self.sid
        else:
            self.logger.warning('Warning: No sessionid please append sessionid: <sessionid> to params before sending.')
        # format url
        url = ("https://" + self.system + "/actifio/api/" + call + "/" + action)
        return method, url, params



from builtins import dict
import numbers
import logging
import six

from actappliance.act_errors import act_errors, ACTError

"""
This module contains the primary objects that power actappliance.
"""

logger = logging.getLogger('models')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


class ActResponse(dict):
    """Dictionary class that helps parsing and raising errors on public Actifio API responses."""

    def __init__(self, *arg, **kwarg):
        # py2 compliance optional argument 'actifio_command'
        self.command = kwarg.pop('actifio_command', None)
        super(ActResponse, self).__init__(*arg, **kwarg)
        self.errorcode = None
        self.errormessage = None

    def raise_for_error(self):
        """Raise ACTError if one occurred"""
        try:
            self.errorcode = self['errorcode']
        except KeyError:
            logger.debug('No errorcode found in ActResponse.')

        try:
            self.errormessage = self['errormessage']
        except KeyError:
            # Handle UI responses
            try:
                self.errormessage = self['msg']
            except KeyError:
                logger.debug('No errormessage found in ActResponse.')

        if self.errorcode or self.errormessage:
            if self.errorcode in act_errors:
                raise act_errors[self.errorcode](self)
            else:
                raise ACTError(self)

    def parse(self, k=None, m_k=None, m_v=None, index=0, warnings=True):
        """
        Takes appliance api response and returns a desired value or dictionary.

        It will first search outside the result for the key. Then it will search inside the result.
        If only a key is given it will return the value of that key from the desired index (default 0).
        If given a match_key and match_value it will search for the first matching dictionary inside the result list
        and return the key from that dictionary.

        Index is ignored without warning if match_key and match_value are provided.

        It is outside the scope of this method to return multiple dictionaries. Parsing multiple outputs shouldn't use
        this method.

        :param self: python object response from appliance api (most commonly json decoded from method cmd)
        :param k: the key; corresponding value will be returned
        :param m_k: match key to find the indexes
        :param m_v: match value to find the indexes
        :param index: index of the list of dictionaries to return
        :param warnings: boolean of whether to raise warning or not
        :return: string if key provided; dictionary if key is not provided
        """
        # Is the result empty?
        try:
            if not self['result']:
                value = self._get_external_value(k)
                if value:
                    logger.info("External value {} found, but will not return from parse.".format(value))
                if warnings:
                    logger.warning('No result found. Returning nothing from parse.')
                return

        # safety net against strange responses
        except (IndexError, TypeError):
            logger.exception('Are you sure you wanted to parse this? You gave no key and there is no result.')
            raise
        except KeyError:
            logger.info('No result dictionary found.')
            return

        # Is the result just an id or response?
        if isinstance(self['result'], (six.string_types, numbers.Integral)):
            logger.debug('result looks like an id or name; returning immediately')
            return self['result']

        # Is the result just info on a single object?
        if isinstance(self['result'], dict):
            if k is None:
                return self['result']
            else:
                return self['result'][k]

        # Is the result several objects?
        if isinstance(self['result'], list):
            # get matching indexes
            if m_k is not None and m_v is not None:
                indexes = []
                for i, dictionary in enumerate(self['result']):
                    if dictionary.get(m_k) == m_v:
                        indexes.append(i)
                if not indexes:
                    logger.info('Nothing found matching {} and {}'.format(m_k, m_v))
                    return
                index = indexes[0]
            if k is None:
                # return dictionary
                return self['result'][index]
            else:
                # Return value
                try:
                    return self.get('result')[index][k]
                except IndexError:
                    logger.error("Index {} doesn't exist. If you manually entered an index, try using matching key "
                                 "and value instead.".format(index))
                    raise
                except KeyError:
                    return self._get_external_value(k)

    def _get_external_value(self, k):
        # Is the response outside of result?
        if k is not None:
            if k in self:
                value = self[k]
                if k in ['status', 'result']:
                    logger.info('There is a top level matching key {}. Using this variable is not a supported use of '
                                'parse.'.format(k))
                    logger.debug('Top level key {} is: {}.'.format(k, value))
                    return
                logger.debug("The top level value for {} is: {}".format(k, value))
                return value
        return



from __future__ import print_function, division, unicode_literals, absolute_import

# Hoist classes and functions into connection namespace
from actappliance.connections.rest import ApplianceRest  # NOQA
from actappliance.connections.ssh import ApplianceSsh    # NOQA

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



import logging
import paramiko
import re
import atexit
from collections import namedtuple
from actappliance.connections.connection import ApplianceConnection


ssh_out = namedtuple('ssh_output', ['out', 'err', 'status'])


class ApplianceSsh(ApplianceConnection):

    def __init__(self, system, connect_args, call_timeout=6 * 60 * 60):
        super(ApplianceSsh, self).__init__()
        self.logger = logging.getLogger(__name__)

        self.system = system
        self.connect_args = connect_args if connect_args is not None else {}
        self.timeout = call_timeout
        # SSH delimiter to use during uds commands
        self.SSH_DELIM = '|MiLeD|'

        # set in method 'connect'
        self.ssh_client = None

    def connect(self):
        """
        Connect to an appliance, disable timeouts and return the ssh object.
        :return: paramiko ssh object
        """
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            ssh.connect(self.system, **self.connect_args)
        except paramiko.SSHException:
            self.logger.exception("Failed to connect to {}".format(self.system))
            raise
        _, out, _ = ssh.exec_command("unset TMOUT")
        rc = out.channel.recv_exit_status()
        if rc != 0:
            raise RuntimeError("Return code of 'unset TMOUT was {}.".format(rc))
        self.ssh_client = ssh

        # Close this connection after execution
        atexit.register(self.disconnect)

    def disconnect(self):
        try:
            self.ssh_client.close()
        except (AttributeError, NameError):
            logging.debug("No ssh client to close.")
        self.ssh_client = None

    def prep_cmd(self, operation, **update_cmds):
        """
        :return: ActCmd object with ssh_command attribute
        """
        prepared_command = self.ActCmd(operation, **update_cmds)
        prepared_command.ssh_command = self.params_to_cli(operation, **update_cmds)
        return prepared_command

    def call(self, command_obj):
        """
        Use when you explicitly want to connect over ssh and send raw ssh commands.

        This houses the functionality unlike rest method because I want to avoid passing around the client to retain
        channels.

        :param command_obj: Instance of ActCmd object
        :return: stdout, stderr, exit status
        """
        if not self.ssh_client:
            self.connect()

        stdin_chan, stdout_chan, stderr_chan = self.ssh_client.exec_command(command_obj.operation)
        stdin_chan.close()

        # Wait for exit_status
        exit_status = stdout_chan.channel.recv_exit_status()

        # get output from channelfiles
        stdout = stdout_chan.readlines()
        stderr = stderr_chan.readlines()
        self.logger.debug('stderr: \n{}'.format(stderr))
        self.logger.debug("exit status: {}".format(exit_status))
        command_obj.call_result = ssh_out(out=stdout, err=stderr, status=exit_status)

    def post_cmd(self, command_obj):
        command_obj.result = self.convert_ssh_output(command_obj.pre.out, command_obj.pre.err)
        return super(ApplianceSsh, self).post_cmd(command_obj)

    def raw(self, ssh_command, statuses=None):
        """
        Send an ssh command with the only processing being newline stripping.

        Threaded timeouts still apply to `raw`. If you need absolute individual control use method `call`.

        :param ssh_command: command to send
        :param statuses: list of acceptable exit statuses
        :return: namedtuple ssh_output
        """
        c = self.ActCmd(ssh_command)
        self.timeout_call(c)
        # Grab the unprocessed call_result before it reaches post_cmd_handler
        assert c.call_result
        out = [line.strip('\n') for line in c.call_result.out]
        err = [line.strip('\n') for line in c.call_result.err]
        # c.call_result.status -> <ssh/rest encapsulation object>.<named tuple>.<ssh exit status>
        output = ssh_out(out=out, err=err, status=c.call_result.status)

        if statuses and output.status not in statuses:
            raise RuntimeError("Exit status of command {} was {}; only {} accepted.".format(
                ssh_command, c.call_result.status, ','.join(str(s) for s in statuses))
            )

        return output

    def params_to_cli(self, operation, **update_cmds):
        """
        Takes a functional cli operation and appends rest like inputs.

        ex.
          self.cmd('udsinfo lshost', filtervalue='ostype=Linux')
        returns:
          "/act/bin/udsinfo lshost -filtervalue 'ostype=Linux'"
        :param operation: A functional cli operation
        :param update_cmds:
        :return:
        """
        cli_command = operation.partition(' ')[0]
        if 'udsinfo' in cli_command:
            cli_command = '/act/bin/udsinfo'
        elif 'udstask' in cli_command:
            cli_command = '/act/bin/udstask'
        elif 'sainfo' in cli_command:
            cli_command = '/act/bin/sainfo'
        else:
            raise ValueError("Actifio appliance doesn't support command {}, accepted values "
                             "(udsinfo, udstask, sainfo)".format(cli_command))
        body = operation.partition(' ')[2]

        # add parameters to body
        for k, v in update_cmds.items():
            # don't pass None params
            if v is not None:
                parameter = "-{}".format(k)
                # TODO: standardize actual string true responses from Rest switch true
                # handle parameterless switches
                if v == 'true':
                    value = ''
                else:
                    value = " '{}'".format(v)
                option = "{}{}".format(parameter, value)
                body += ' {}'.format(option)
        if 'udsinfo' in cli_command and body.startswith(('ls', 'list')):
            body += " -delim '{}'".format(self.SSH_DELIM)
        full_operation = "{} {}".format(cli_command, body)
        return full_operation

    def convert_ssh_output(self, stdout, stderr):
        self.logger.debug('Parsing stdout: {}'.format(stdout))
        rest_like_result = {}
        # First format stderr
        if stderr:
            error_code = re.match("(?:ACTERR-)(\d+)", stderr[0]).group(1)
            error_message = re.match("(?:ACTERR-\d+ )(.+)", stderr[0]).group(1)
            rest_like_result['errorcode'] = error_code
            rest_like_result['errormessage'] = error_message
        # Now format stdout
        try:
            # Handle udsinfo commands that start with ls
            if self.SSH_DELIM in stdout[0]:
                self.logger.info('Ssh operation contains delim, grabbing header.')
                headers = stdout[0].split(self.SSH_DELIM)
                values = [row.split(self.SSH_DELIM) for row in stdout[1:]]
                # Make the result list like JSON response
                result_list = []
                for row in values:
                    result_list.append(dict(zip(headers, row)))
                rest_like_result['result'] = result_list
            else:
                # Handle all other udsinfo commands
                # only return a list if there is more than one item, else return a string (see exception)
                if stdout[1]:
                    rest_like_result = {'result': stdout}
                # TODO Handle commands that match "udsinfo -h | egrep -v '^\s*ls'"
        except IndexError:
            try:
                # return single item response as a string
                rest_like_result['result'] = stdout[0]
            except IndexError:
                # return "[]" if stdout was empty, matching standard rest response
                rest_like_result['result'] = stdout
        self.logger.debug("Converted cli result: {}".format(rest_like_result))
        return rest_like_result



"""
Create an act_errors dictionary with the values being  Actifio exceptions. The use case being you can catch individual
exceptions instead of making comparisons on the caught generic exception, without requiring star imports. Also, most
users are more familiar with error codes than exception names, so referencing with a number is better.

Allows this logic:
    except act_errors[10016]:
        <handle exception>
    except act_errors[8675309]:
        <handle Jenny>

to replace this:
    except ACTError:
        if int(e.errorcode) == 10016:
            <handle exception>
        elif int(e.errorcode) == 8675309:
            <handle Jenny>
        else:
            raise

This also improves the legibility of handling several possible ACTErrors of a try block.
"""
from actappliance.misc import camelize


class ACTError(Exception):

    def __init__(self, response_obj, *args, **kwargs):
        super(ACTError, self).__init__(*args, **kwargs)
        self.errormessage = response_obj.errormessage
        self.errorcode = response_obj.errorcode
        self.msg = '{}: {}'.format(self.errorcode, self.errormessage)

    def __str__(self):
        return self.msg


_error_codes = {
    'BAD_NUMBER_FORMAT': 10001,
    'FILE_NOT_EXIST': 10002,
    'FILE_EXISTS': 10003,
    'UNKNOWN_TYPE': 10004,
    'MISSING_ARGUMENT': 10005,
    'VIOLATE_UNIQUENESS': 10006,
    'STILL_IN_USE': 10007,
    'INVALID_VALUE': 10008,
    'PROTECTED_SYSTEM_RESOURCE': 10009,
    'INVALID_NAME': 10010,
    'LOGIN_FAILURE': 10011,
    'DISABLE_SNAP': 10012,
    'DISABLE_DEDUP': 10013,
    'EXCEEDED_LIMIT': 10014,
    'CLUSTER_CONNECTION_FAILURE': 10015,
    'OBJECT_NOT_EXIST': 10016,
    'UNKNOWN_OBJECT': 10016,
    'INVALID_OPERATION': 10017,
    'INVALID_OPTION': 10018,
    'SYSTEM_RESOURCE': 10019,
    'INVALID_SESSION': 10020,
    'INVALID_FILTERVALUE': 10021,
    'CONNECTION_FAILURE': 10022,
    'OPERATION_FAILURE': 10023,
    'SECURITY_VIOLATION': 10024,
    'MDISK_OFFLINE': 10025,
    'SYSTEM_ERROR': 10026,

    'EXCEED_DED_LIMIT': 10031,
    'EXCEED_PER_LIMIT': 10032,
    'EXCEED_PRI_LIMIT': 10033,
    'EXCEED_FC_LIMIT': 10034,
    'EXCEED_RC_LIMIT': 10035,
    'EXCEED_COPY_LIMIT': 10036,
    'EXCEED_MIRROR_LIMIT': 10037,
    'EXCEED_VDISK_LIMIT': 10038,
    'PING_TIMEOUT': 10039,
    'BAD_ARGUMENT': 10040,
    'DATA_OUTOFRANGE': 10041,
    'SOCKET_TIMEOUT': 10042,
    'SLA_VIOLATION': 10043,
    'NETWORK_INTERFACE_FAILURE': 10044,
    'SAFE_DED_LIMIT': 10045,
    'SAFE_PER_LIMIT': 10046,
    'SAFE_PRI_LIMIT': 10047,
    'MIN_DED_LIMIT_XING': 10048,
    'EXCEED_MIN_DED_LIMIT': 10049,

    'SECURITY_TURNON': 10050,
    'SECURITY_TURNOFF': 10051,
    'UPDATE_IN_PROCESS': 10052,
    'OPERATION_TIMEOUT': 10053,
    'INVALID_XML': 10054,
    'REMOTE_PROTECTION_OFFLINE': 10055,
    'NODE_OFFLINE': 10056,
    'NODE_ASSERT': 10057,

    'DEDUP_DOWN': 10060,
    'DEDUP_DISABLED': 10061,
    'MULTIPLE_VMS_SAME_NAME': 10062,
    'DNS_LOOKUP_FAILED': 10063,
    'SSD_EXCEED_WEAR_LEVEL': 10064,

    'DISABLE_UDPPM_SCHEDULER': 10070,
    'INVALID_SVC_MIRROR': 10071,
    'EXCEEDING_MDL': 10072,

    'CLUSTER_INVALID_OPERATION': 10080,
    'CLUSTER_C[2]C_FAILURE': 10081,

    'LICENSE_VIOLATION': 10082,
}

# Create a dictionary of errors
act_errors = {}
for error in _error_codes:
    actifio_error = camelize(error.lower()) + 'Error'
    err_num = _error_codes[error]
    docstring = "A {0} {1} Actifio error.".format(err_num, error)
    cls_dict = {'__doc__': docstring}
    act_errors[err_num] = (type(actifio_error, (ACTError,), cls_dict))



import logging
from actappliance.act_errors import act_errors, ACTError


class CLIError(ACTError):
    """A CLI error occurred."""


class RESTError(ACTError):
    """A REST error occurred"""


class ApplianceError(Exception):
    """Generic Appliance exception to raise and log different fatal errors."""

    def __init__(self, msg):
        super(ApplianceError).__init__(type(self))
        self.msg = "E: %s" % msg

    def __str__(self):
        return self.msg

    def __unicode__(self):
        return self.msg


def retry_if_result_none(result):
    """Return True if we should retry (in this case when result is None), False otherwise"""
    return result is None


def retry_if_result_not_none(result):
    """Return False if we should retry (in this case when result isn't None), False otherwise"""
    return result is not None


def retry_if_act_result_empty(act_response_obj):
    """Return True if result is populated"""
    try:
        ok = act_response_obj['result'] == []
    except KeyError:
        # Retry if result doesn't exist. This allows the call to continue until it functions
        # Primary use is to search job history until the result is available
        logging.debug('No result found in ActResponse, flagging for retry.')
        ok = True
    except TypeError:
        logging.exception("It is very likely your function doesn't have a return where it should.")
        raise
    return ok


def retry_if_act_result_populated(act_response_obj):
    """
    Return True if result is empty/falsey.
    This will raise error if result doesn't exist
    """
    try:
        if act_response_obj['result']:
            ok = True
        else:
            ok = False
    except KeyError:
        logging.debug('No result found in ActResponse, reraising.')
        # If this is ever changed must return False as retries will happen if return is None
        raise
    except TypeError:
        logging.exception("It is very likely your function doesn't have a return where it should.")
    return ok


def retry_if_acterror(exception):
    """Return True if we should retry (in this case when it's an ACTError or subclass of ACTError), False otherwise"""
    logging.debug("Exception passed to retry_if_acterror is: {}".format(exception))
    # The exception is the instance
    return issubclass(exception.__class__, ACTError)


def retry_if_no_acterror(exception):
    """Opposite of retry_if_acterror"""
    return not issubclass(exception.__class__, ACTError)


def retry_if_applianceerror(exception):
    """Retry if ApplianceError occurs"""
    return isinstance(exception, ApplianceError)


def retry_if_unknownobject(exception):
    return isinstance(exception, act_errors[10016])


def retry_until_unknownobject(exception):
    return not isinstance(exception, act_errors[10016])


def retry_if_workflow_is_deleted(exception):
    """**Deprecated** please use retry_until_unknownobject"""
    return not isinstance(exception, act_errors[10016])


def retry_if_workflow_is_running(act_response_obj):
    """Retry unless ActResponse['result'] is populated and doesn't contain 'RUNNING'"""
    # retry if empty result
    ok = True
    if retry_if_act_result_populated(act_response_obj):
        if 'RUNNING' not in act_response_obj['result']:
            ok = False
    return ok


def retry_until_runtime(exception):
    return isinstance(exception, RuntimeError)
